from ..db import model as m

def insert_study(record: m.SraStudy):
    def insert_attributes():
        def insert_tags():
    def insert_identifiers():
        def insert_namespace()
    # study_accession (FK to study)
    # identifier(text)
    # namespace_id (
    # TODO: need to convert namespace to use id
    
    def insert_pubmed_ids()
    def 



#!/usr/bin/env python
import argparse
import omicidx.sra_parsers
import json
import logging
import collections
from xml.etree import ElementTree as et



def main():
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(description="""
    Transforms an SRA XML mirroring metadata file into
    corresponding JSON format files. JSON is line-delimited
    JSON (not an array).""")
    
    parser.add_argument('fname',
                        help = "The SRA xml filename")
    parser.add_argument('outfile',
                        help="the output filename where to write json records")

    opts = parser.parse_args()

    entity=""

    if('study' in opts.fname):
        entity = "STUDY"
        sra_parser = omicidx.sra_parsers.SRAStudyRecord

    if('run' in opts.fname):
        entity = "RUN"
        sra_parser = omicidx.sra_parsers.SRARunRecord

    if('sample' in opts.fname):
        entity = "SAMPLE"
        sra_parser = omicidx.sra_parsers.SRASampleRecord

    if('experiment' in opts.fname):
        entity = "EXPERIMENT"
        sra_parser = omicidx.sra_parsers.SRAExperimentRecord

    logger.info('configuration: {}'.format(str(opts)))
    logger.info('using {} entity type'.format(entity))
    logger.info('parsing {} records'.format(entity))
    n = 0
    with open(opts.outfile, 'w') as outfile:
        with omicidx.sra_parsers.open_file(opts.fname) as f:
            for event, element in et.iterparse(f):
                if(event == 'end' and element.tag == entity):
                    rec = sra_parser(element).data
                    n+=1
                    if((n % 100000)==0):
                        logger.info('parsed {} {} entries'.format(entity, n))
                    outfile.write(json.dumps(rec) + "\n")
                    element.clear()
            logger.info('parsed {} entity entries'.format(n))

if __name__ == '__main__':
    main()



# alembic
# need to create schema by hand:
# upgrade:     op.execute("create schema etl")
# downgrade:     op.execute("drop schema etl")
#  - alembic revision --autogenerate -m "Added SRA study table skeleton"
#  - alembic upgrade head
#  - alembic history
#  - alembic downgrade HASH
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql.json import JSONB
from sqlalchemy import (Column, Integer, String, UniqueConstraint,
                        ForeignKeyConstraint, Table, Text, ForeignKey,
                        PrimaryKeyConstraint, Boolean, DateTime, Numeric, text,
                        Index)
from sqlalchemy.orm import (relationship)

Base = declarative_base()


class SraStudyJson(Base):
    __tablename__ = 'sra_study_json'
    accession = Column(String(20), primary_key=True)
    doc = Column(JSONB)


class SraSampleJson(Base):
    __tablename__ = 'sra_sample_json'
    accession = Column(String(20), primary_key=True)
    doc = Column(JSONB)


class SraExperimentJson(Base):
    __tablename__ = 'sra_experiment_json'
    accession = Column(String(20), primary_key=True)
    doc = Column(JSONB)


class SraRunJson(Base):
    __tablename__ = 'sra_run_json'
    accession = Column(String(20), primary_key=True)
    doc = Column(JSONB)


##########
# MIXINS #
##########


class SraNamespace(Base):
    __tablename__ = 'sra_namespace'

    namespace = Column(Text, primary_key=True)
    identifiers = relationship('SraSampleIdentifier')


class SraStudyIdentifier(Base):
    __tablename__ = 'sra_study_identifier'

    study_accession = Column(ForeignKey('sra_study.accession'),
                             primary_key=True)
    study = relationship('SraStudy', back_populates='identifiers')
    identifier = Column(Text, primary_key=True)
    namespace = Column(ForeignKey('sra_namespace.namespace'), primary_key=True)


class SraStudy(Base):
    __tablename__ = 'sra_study'

    abstract = Column(Text)
    BioProject = Column(String(30))
    Geo = Column(String(15), index=True)
    accession = Column(String(20), primary_key=True)
    center_name = Column(Text)  # needs to be pulled out
    broker_name = Column(Text)  # needs to be pulled out
    description = Column(Text, comment="full description of study")
    study_type = Column(Text)  # needs to be pulled out
    title = Column(Text)
    received = Column(DateTime)
    updated = Column(DateTime)
    published = Column(DateTime)
    status = Column(Text)  # needs to be pulled out
    insdc = Column(Boolean)
    identifiers = relationship('SraStudyIdentifier', back_populates='study')


class LibrarySource(Base):
    __tablename__ = 'library_source'

    id = Column(Integer, primary_key=True)
    value = Column(String)


class LibraryStrategy(Base):
    __tablename__ = 'library_strategy'

    id = Column(Integer, primary_key=True)
    value = Column(String)


class LibrarySelection(Base):
    __tablename__ = 'library_selection'

    id = Column(Integer, primary_key=True)
    value = Column(String)


class LibraryLayout(Base):
    __tablename__ = 'library_layout'

    id = Column(Integer, primary_key=True)
    value = Column(String)


class CenterName(Base):
    __tablename__ = 'center_name'

    id = Column(Integer, primary_key=True)
    value = Column(String)


class InstrumentModel(Base):
    __tablename__ = 'instrument_model'

    id = Column(Integer, primary_key=True)
    value = Column(String)


class Platform(Base):
    __tablename__ = 'platform'

    id = Column(Integer, primary_key=True)
    value = Column(String)


def to_tsvector_ix(*columns):
    s = " || ' ' || ".join(columns)
    return func.to_tsvector('english', text(s))


from sqlalchemy.sql import func


class SraExperimentIdentifier(Base):
    __tablename__ = 'sra_experiment_identifier'

    experiment_accession = Column(ForeignKey('sra_experiment.accession'),
                                  primary_key=True)
    experiment = relationship('SraExperiment', back_populates='identifiers')
    identifier = Column(Text, primary_key=True)
    namespace = Column(ForeignKey('sra_namespace.namespace'), primary_key=True)


class SraExperiment(Base):
    __tablename__ = 'sra_experiment'

    accession = Column(String(20), primary_key=True)
    alias = Column(String)
    # attributes
    center_name_id = Column(ForeignKey('center_name.id'), index=True)
    design = Column(Text)
    description = Column(Text)
    identifiers = relationship('SraExperimentIdentifier',
                               back_populates='experiment')
    instrument_model = Column(ForeignKey('instrument_model.id'), index=True)
    library_name = Column(Text)
    library_construction_protocol = Column(Text)
    library_layout_orientation = Column(Text)  #appears to always be NULL
    library_layout_length = Column(Numeric)
    library_layout_sdev = Column(Numeric)
    library_selection_id = Column(Integer,
                                  ForeignKey('library_selection.id'),
                                  index=True)
    library_strategy_id = Column(Integer,
                                 ForeignKey('library_strategy.id'),
                                 index=True)
    library_source_id = Column(Integer,
                               ForeignKey('library_source.id'),
                               index=True)
    library_layout_id = Column(Integer,
                               ForeignKey('library_layout.id'),
                               index=True)
    platform_id = Column(Integer,
                         ForeignKey('platform.id'),
                         index=True,
                         comment='The sequencing platform')
    # xreefs
    sample_accession = Column(String(20),
                              ForeignKey('sra_sample.accession'),
                              index=True)
    study_accession = Column(String(20),
                             ForeignKey('sra_study.accession'),
                             index=True)
    title = Column(Text)
    # ?visibility
    # ?replacedby

    # TODO: Need to figure out how to add this with alembic
    __table_args__ = (Index('ix_fts_experiment',
                            to_tsvector_ix('title', 'description', 'design',
                                           'alias', 'accession'),
                            postgresql_using='gin'), )


class SraSampleIdentifier(Base):
    __tablename__ = 'sra_sample_identifier'

    sample_accession = Column(ForeignKey('sra_sample.accession'),
                              primary_key=True)
    sample = relationship('SraSample', back_populates='identifiers')
    identifier = Column(Text, primary_key=True)
    namespace = Column(ForeignKey('sra_namespace.namespace'), primary_key=True)


class SraSample(Base):
    __tablename__ = 'sra_sample'

    accession = Column(String(20), primary_key=True)
    # TODO: foreign key to geo
    geo = Column(String(15), index=True)
    # TODO: foreigh key to biosample
    BioSample = Column(String, index=True)
    title = Column(Text)
    alias = Column(Text)
    experiments = relationship('SraExperiment', backref='sample')
    # leaving out organism
    # TODO: Taxonomy link
    taxon_id = Column(ForeignKey('taxonomy.id'), index=True)
    description = Column(Text)
    identifiers = relationship('SraSampleIdentifier', back_populates='sample')
    # attributes
    # xrefs
    # link to experiment


class BaseCounts(Base):
    __tablename__ = 'base_counts'

    sra_run_accession = Column(ForeignKey('sra_run.accession'),
                               index=True,
                               primary_key=True)
    base = Column(String(1), primary_key=True)
    count = Column(Integer)


class BaseQualities(Base):
    __tablename__ = 'base_quality'

    sra_run_accession = Column(ForeignKey('sra_run.accession'),
                               index=True,
                               primary_key=True)
    quality = Column(Integer, primary_key=True)
    count = Column(Integer)


class RunRead(Base):
    __tablename__ = 'run_read'

    sra_run_accession = Column(ForeignKey('sra_run.accession'),
                               index=True,
                               primary_key=True)
    index = Column(Integer, primary_key=True)
    mean_length = Column(Numeric)
    sd_length = Column(Numeric)


class RunFileAlternative(Base):
    __tablename__ = 'run_file_alternative'
    __table_args__ = (UniqueConstraint('run_fileset_id', 'url'), )

    id = Column(Integer, primary_key=True)
    run_fileset_id = Column(ForeignKey('run_fileset.id'), index=True)
    url = Column(String)
    free_egress = Column(String)  # TODO: consider foreign key
    access_type = Column(String)  # TODO: consider foreign key
    org = Column(String, index=True)


class RunFileSet(Base):
    __tablename__ = 'run_fileset'
    __table_args__ = (UniqueConstraint('run_accession', 'filename'), )

    id = Column(Integer, primary_key=True)
    run_accession = Column(String(20), ForeignKey('sra_run.accession'))
    filename = Column(String)
    url = Column(String)
    size = Column(Integer)
    date = Column(DateTime)
    md5 = Column(String)
    alternatives = relationship('run_file_alternative', backref='run_fileset')
    sratoolkit = Column(String)


class SraRun(Base):
    __tablename__ = 'sra_run'

    accession = Column(String(20), primary_key=True)
    alias = Column(String)
    run_date = Column(DateTime, index=True)
    run_center = Column(String)
    center_name_id = Column(ForeignKey('center_name.id'), index=True)
    total_spots = Column(Integer, index=True)
    total_bases = Column(Integer, index=True)
    size = Column(Integer)
    load_done = Column(Boolean)
    published = Column(DateTime)
    is_public = Column(Boolean)
    cluster_name = Column(String)
    avg_length = Column(Numeric, index=True)
    experiment_accession = Column(String(20),
                                  ForeignKey('sra_experiment.accession'),
                                  index=True, nullable=False)
    experiment = relationship('SraExperiment', back_populates='runs')
    # attributes
    # files
    # qualities
    # basecounts
    # reads


class TaxonCountAnalysis(Base):
    __tablename__ = 'taxon_count_analysis'

    id = Column(Integer, primary_key=True)
    nspot_analyze = Column(Integer, index=True)
    total_spots = Column(Integer, index=True)
    run_accession = Column(ForeignKey("sra_run.accession"), index=True)
    run = relationship('SraRun', backref='taxon_analysis')


class Taxonomy(Base):
    __tablename__ = "taxonomy"

    id = Column(Integer, primary_key=True)
    rank = Column(String, index=True)  # ? normalize further?
    name = Column(String, index=True)  # ? unique
    parent = Column(Integer, ForeignKey('taxonomy.id'), index=True)


class TaxonCountEntry(Base):
    __tablename__ = 'taxon_count_entry'

    taxon_analysis_id = Column(Integer,
                               ForeignKey('taxon_count_analysis.id'),
                               primary_key=True)
    taxon_id = Column(Integer, ForeignKey('taxonomy.id'), primary_key=True)
    self_count = Column(Integer)
    total_count = Column(Integer)
    taxon_analysis = relationship('TaxonCountAnalysis', backref='taxon_counts')


geo_series_contributors = Table(
    'geo_series_contributors', Base.metadata,
    Column('gse_accession',
           ForeignKey('geo_series.accession'),
           primary_key=True),
    Column('geo_name_id', ForeignKey('geo_name.id'), primary_key=True))

geo_sample_contributors = Table(
    'geo_sample_contributors', Base.metadata,
    Column('gsm_accession',
           ForeignKey('geo_sample.accession'),
           primary_key=True),
    Column('geo_name_id', ForeignKey('geo_name.id'), primary_key=True))


class GeoName(Base):
    __tablename__ = 'geo_name'
    __table_args__ = (UniqueConstraint('first_name',
                                       'middle_name',
                                       'last_name',
                                       name='ix_uq_geo_name'), )

    id = Column(Integer, primary_key=True)
    first_name = Column(String)
    middle_name = Column(String)
    last_name = Column(String)

    series = relationship('GEOSeries', secondary=geo_series_contributors)


class GeoContact(Base):
    __tablename__ = 'geo_contact'

    id = Column(Integer, primary_key=True)
    name_id = Column(ForeignKey('geo_name.id'), index=True)
    email = Column(String)
    state = Column(String)
    address = Column(String)
    department = Column(String)
    country = Column(String)
    web_link = Column(String)
    institute = Column(String)
    zip_postal_code = Column(String)
    phone = Column(String)


class GeoSeriesType(Base):
    __tablename__ = 'geo_series_type'

    id = Column(Integer, primary_key=True)
    value = Column(String, unique=True)


class GeoCharacterisricTag(Base):
    __tablename__ = 'geo_characteristic_tag'

    id = Column(Integer, primary_key=True)
    tag = Column(String, unique=True)
    
class GeoCharacteristic(Base):
    __tablename__ = 'geo_characteristic'

    id = Column(Integer, primary_key=True)
    tag_id = Column(ForeignKey('geo_characteristic_tag.id'))
    value = Column(String)
    
    
class GeoSample(Base):
    __tablename__ = 'geo_sample'

    accession = Column(String(15), primary_key=True)
    type = Column(String, index=True)
    biosample = Column(String, index=True)  # TODO: convert to foreign key?
    tag_count = Column(Integer)
    tag_length = Column(Integer)
    platform_accession = Column(String(15), index=True)
    hyb_protocol = Column(String)
    channel_count = Column(Integer)
    scan_protocol = Column(String)
    data_row_count = Column(Integer)
    library_source = Column(String)
    overall_design = Column(String)
    sra_experiment = Column(String)  # TODO: convert to foreign keys
    data_processing = Column(String)
    contributors = relationship('GeoName', secondary=geo_sample_contributors)


class GeoSeries(Base):
    __tablename__ = 'geo_series'

    accession = Column(String(15), primary_key=True)
    summary = Column(String)
    contributors = relationship('GeoName', secondary=geo_series_contributors)
    # TODO: form relationshop
    # sra_studies = relationship('SraStudy', secondary=geo_series_sra_study)
    # contact ? many-to-one?
    # TODO: type (many-to-many)
    # TODO: pubmed (many-to-many)
    # TODO: samples (many-to-many)
    # TODO: taxid (many-to-many)
    # TODO: platform (many-to-many)
    data_processing = Column(String)
    description = Column(String)
    overall_design = Column(String)

def render_erd(fname = 'omicidx_erd.pdf'):
    """Render the entire schema as a pdf/png based on filename suffix"""
    try: 
        from eralchemy import render_er
        render_er(Base, fname)
    except:
        logging.error('need to install eralchemy for this functionality')



# coding: utf-8
from sqlalchemy import BigInteger, Column, DateTime, ForeignKey, Integer, Table, Text, text, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB, TSVECTOR
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy.dialects.postgresql as postgresql

Base = declarative_base()
metadata = Base.metadata

etl_schema = 'etl'
public_schema = 'public2'
staging_schema = 'staging'

class IdentifiersMixin(object):
    identifier = Column(Text)
    namespace = Column(Text)

class XrefsMixin(object):
    db = Column(Text)
    identifier = Column(Text)
    uuid = Column(Text)

class AttributesMixin(object):
    tag = Column(Text)
    value = Column(Text)



t_biosample_jsonb = Table(
    'biosample_jsonb', metadata,
    Column('doc', JSONB(astext_type=Text())),
    schema=etl_schema
)


t_biosample_test = Table(
    'biosample_test', metadata,
    Column('id', Integer),
    Column('gsm', Text),
    Column('dbgap', Text),
    Column('model', Text),
    Column('title', Text),
    Column('access', Text),
    Column('id_recs', JSONB(astext_type=Text())),
    Column('ids', JSONB(astext_type=Text())),
    Column('taxon_id', Integer),
    Column('accession', Text),
    Column('attributes', JSONB(astext_type=Text())),
    Column('sra_sample', Text),
    Column('description', Text),
    Column('last_update', DateTime),
    Column('publication_date', DateTime),
    Column('submission_date', DateTime),
    Column('is_reference', Text),
    Column('taxonomy_name', Text),
    Column('textsearchable_index_col', TSVECTOR, index=True),
    schema=etl_schema
)


t_experiment_jsonb = Table(
    'experiment_jsonb', metadata,
    Column('doc', JSONB(astext_type=Text())),
    schema=etl_schema
)


t_fileinfo_addons = Table(
    'fileinfo_addons', metadata,
    Column('accession', Text, index=True),
    Column('file_size', BigInteger),
    Column('file_md5', Text),
    Column('file_date', DateTime),
    Column('filename', Text),
    Column('url', Text),
    schema=etl_schema
)


t_fileinfo_runs = Table(
    'fileinfo_runs', metadata,
    Column('accession', Text, index=True),
    Column('file_size', BigInteger),
    Column('file_md5', Text),
    Column('file_date', DateTime),
    schema=etl_schema
)


t_geo_jsonb = Table(
    'geo_jsonb', metadata,
    Column('doc', JSONB(astext_type=Text())),
    Column('update_time', DateTime, server_default=text("'2019-03-20 12:55:47.158602'::timestamp without time zone")),
    schema=etl_schema
)


t_run_jsonb = Table(
    'run_jsonb', metadata,
    Column('doc', JSONB(astext_type=Text())),
    schema=etl_schema
)


t_sample_jsonb = Table(
    'sample_jsonb', metadata,
    Column('doc', JSONB(astext_type=Text())),
    schema=etl_schema
)


class SraAccession(Base):
    __tablename__ = 'sra_accession'
    __table_args__ = {'schema': etl_schema}

    accession = Column(Text, primary_key=True, index=True)
    submission = Column(Text)
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    type = Column(Text)
    center = Column(Text)
    visibility = Column(Text)
    alias = Column(Text)
    experiment = Column(Text)
    sample = Column(Text)
    study = Column(Text)
    loaded = Column(Text)
    spots = Column(BigInteger)
    bases = Column(BigInteger)
    md5sum = Column(Text)
    bio_sample = Column(Text)
    bio_project = Column(Text)
    replaced_by = Column(Text)



class SraStudy(Base):
    __tablename__ = 'sra_study'
    __table_args__ = {'schema': public_schema}

    bioproject = Column(Text)
    gse = Column(Text)
    abstract = Column(Text)
    accession = Column(Text, primary_key=True)
    alias = Column(Text)
    attributes = Column(JSONB(astext_type=Text()))
    broker_name = Column(Text)
    center_name = Column(Text)
    description = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    study_type = Column(Text)
    title = Column(Text)
    xrefs = Column(JSONB(astext_type=Text()))
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    bio_project = Column(Text)
    replaced_by = Column(Text)



class SraSample(Base):
    __tablename__ = 'sra_sample'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    alias = Column(Text)
    attributes = Column(JSONB(astext_type=Text()))
    bio_sample = Column(Text)
    broker_name = Column(Text)
    center_name = Column(Text)
    description = Column(Text)
    gsm = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    organism = Column(Text)
    title = Column(Text)
    taxon_id = Column(Integer)
    xrefs = Column(JSONB(astext_type=Text()))
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    replaced_by = Column(Text)
    study_accession = Column(ForeignKey(public_schema + '.sra_study.accession'))

    sra_study = relationship('SraStudy')

t_study_jsonb = Table(
    'study_jsonb', metadata,
    Column('doc', JSONB(astext_type=Text())),
    schema=etl_schema
)


class SraExperiment(Base):
    __tablename__ = 'sra_experiment'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    alias = Column(Text)
    attributes = Column(JSONB(astext_type=Text()))
    broker_name = Column(Text)
    center_name = Column(Text)
    description = Column(Text)
    design = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    instrument_model = Column(Text)
    library_construction_protocol = Column(Text)
    library_layout_length = Column(Text)
    library_layout_orientation = Column(Text)
    library_layout_sdev = Column(Text)
    library_name = Column(Text)
    library_strategy = Column(ForeignKey(public_schema + '.sra_library_strategy.value'), index=True)
    library_source = Column(ForeignKey(public_schema + '.sra_library_source.value'), index=True)
    library_layout = Column(ForeignKey(public_schema + '.sra_library_layout.value'), index=True)
    library_selection = Column(ForeignKey(public_schema + '.sra_library_selection.value'), index=True)
    platform = Column(ForeignKey(public_schema + '.sra_platform.value'), index=True)
    sample_accession = Column(ForeignKey(public_schema + '.sra_sample.accession'))
    study_accession = Column(ForeignKey(public_schema + '.sra_study.accession'))
    title = Column(Text)
    xrefs = Column(Text)
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    replaced_by = Column(Text)

    sra_sample = relationship('SraSample')
    sra_study = relationship('SraStudy')


class SraRun(Base):
    __tablename__ = 'sra_run'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    broker_name = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    run_center = Column(Text)
    center_name = Column(Text)
    run_date = Column(DateTime)
    attributes = Column(JSONB(astext_type=Text()))
    nreads = Column(Integer)
    alias = Column(Text)
    spot_length = Column(Integer)
    experiment_accession = Column(ForeignKey(public_schema + '.sra_experiment.accession'))
    study_accession = Column(ForeignKey(public_schema + '.sra_study.accession'))
    sample_accession = Column(ForeignKey(public_schema + '.sra_sample.accession'))
    reads = Column(JSONB(astext_type=Text()))
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    bio_project = Column(Text)
    replaced_by = Column(Text)
    loaded = Column(Text)
    spots = Column(BigInteger)
    bases = Column(BigInteger)

    sra_experiment = relationship('SraExperiment')
    sra_study = relationship('SraStudy')
    sra_sample = relationship('SraSample')


class SraLibraryStrategy(Base):
    __tablename__ = 'sra_library_strategy'
    __table_args__ = {'schema': public_schema}
    
    value = Column(Text, primary_key=True)

class SraLibrarySource(Base):
    __tablename__ = 'sra_library_source'
    __table_args__ = {'schema': public_schema}
    
    value = Column(Text, primary_key=True)

class SraLibraryLayout(Base):
    __tablename__ = 'sra_library_layout'
    __table_args__ = {'schema': public_schema}
    
    value = Column(Text, primary_key=True)

class SraLibrarySelection(Base):
    __tablename__ = 'sra_library_selection'
    __table_args__ = {'schema': public_schema}
    
    value = Column(Text, primary_key=True)

class SraPlatform(Base):
    __tablename__ = 'sra_platform'
    __table_args__ = {'schema': public_schema}
    
    value = Column(Text, primary_key=True)



#########################
# GEO models below here #
#########################


geo_gse_gsm = Table('geo_gse_gsm', Base.metadata,
    Column('gse', Text, ForeignKey('geo_gse.accession')),
    Column('gsm', Text, ForeignKey('geo_gsm.accession'))
)


geo_gse_gpl = Table('geo_gse_gpl', Base.metadata,
    Column('gse', Text, ForeignKey('geo_gse.accession')),
    Column('gpl', Text, ForeignKey('geo_gpl.accession'))
)


class GEOSeries(Base):
    __tablename__ = 'geo_gse'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    title = Column(Text)
    status = Column(Text)
    summary = Column(Text)
    description = Column(Text)
    overall_design = Column(Text)
    submission_date = Column(DateTime)
    last_update_date = Column(DateTime)
    data_processing = Column(Text)
    contributors = Column(postgresql.ARRAY(Text))
    gpls = relationship("GEOPlatform",
                        secondary = geo_gse_gpl,
                        back_populates = "gses")
    gsms = relationship("GEOSample",
                        secondary = geo_gse_gsm,
                        back_populates = "gses")


class GEOPlatform(Base):
    __tablename__ = 'geo_gpl'
    __table_args__ = {'schema': public_schema}
    
    accession = Column(Text, primary_key=True)
    title = Column(Text)
    status = Column(Text)
    summary = Column(Text)
    description = Column(Text)
    submission_date = Column(DateTime)
    last_update_date = Column(DateTime)
    technology = Column(Text)
    distribution = Column(Text)
    data_row_count = Column(Integer)
    contributors = Column(postgresql.ARRAY(Text))
    gses = relationship("GEOSeries",
                        secondary = geo_gse_gpl,
                        back_populates = "gpls")
    gsms = relationship("GEOSample",
                        secondary = geo_gse_gsm,
                        back_populates = "gpls")


    
class GEOContact(Base):
    __tablename__ = 'geo_contact'
    __table_args__ = {'schema': public_schema}

    email = Column(Text, primary_key=True)
    name = Column(Text)
    phone = Column(Text)
    fax = Column(Text)
    city = Column(Text)
    address = Column(Text)
    country = Column(Text)
    web_link = Column(Text)
    institute = Column(Text)
    postal_code = Column(Text)


biosample_to_attributes = Table(
    'biosample_to_attributes', Base.metadata,
    Column('id', Integer, primary_key=True),
    Column('biosample_id', ForeignKey('biosample.id')),
    Column('biosample_attribute_id', ForeignKey('biosample_attribute.id')),
)

class BiosampleModel(Base):
    __tablename__ = "biosample_model"

    id = Column(Integer, primary_key=True)
    name = Column(Text, index=True)


class Biosample(Base):
    __tablename__ = "biosample"

    id = Column(BigInteger, primary_key=True)
    submission_date = Column(DateTime, index=True, nullable=False)
    last_update = Column(DateTime, index=True, nullable=False)
    publication_date = Column(DateTime, index=True, nullable=False)
    access = Column(Text, index=True, comment="access control, public or controlled")
    # srs = Column(ForeignKey('sra_sample.accession'), index=True)
    # gsm = Column(ForeignKey('geo_gsm.accession'), index=True)
    dbgap = Column(Text, index=True, comment = "dbGaP identifier")
    title = Column(Text)
    description = Column(Text)
    taxonomy_name = Column(Text)
    taxon_id = Column(Integer, index=True, comment = "NCBI Taxonomy id")
    model = Column(ForeignKey("biosample_model.id"), index=True)
    attributes = relationship("BiosampleAttribute", secondary = biosample_to_attributes)
                                
class BiosampleAttribute(Base):
    __tablename__ = "biosample_attribute"

    id = Column(Integer, primary_key=True)
    name = Column(Text, index=True)
    harmonized_name = Column(Text, index = True)
    display_name = Column(Text, index=True)
    value = Column(Text, index=True)
    UniqueConstraint('name', 'harmonized_name', 'value')
    biosamples = relationship("Biosample", secondary = biosample_to_attributes)


class BiosampleIdentifier(Base):
    __tablename__ = "biosample_identifier"

    id = Column(Integer, primary_key=True)
    biosample = Column(ForeignKey('biosample.id'), index=True)
    db = Column(Text, index=True)
    label = Column(Text, index = True)
    identifier = Column(Text, index=True)



"""Biosample parser

Implemented as an iterator


>>> import omicidx.biosample as b
>>> for bios in b.BioSampleParser('biosample_set.xml.gz'):
>>>     print(bios.as_json())
"""

import os.path
import xml.etree.ElementTree as ET
import gzip
import json
import logging
import click
import subprocess

class BioSample(dict):
    """BioSample class"""
    def __init__(self):
        self.update()
    
    def as_json(self,indent=None):
        return(json.dumps(self))

    def __str__(self):
        return('{0}'.format(self['title']))

    def __repr__(self):
        return('BioSample <id={0} title={1}>'.format(self['id'],self['title']))
    
class BioSampleParser(object):
    """Parse a BioSample xml file.

    Implemented as an iterator"""

    
    def __init__(self,fname):
        """Init

        Handles .gz as well as plain text"""
        self.fhandle = None
        if(fname.endswith('gz')):
            self.fhandle = gzip.open(fname)
        else:
            self.fhandle = open(fname)
        self.context = ET.iterparse(self.fhandle, events=("start", "end"))
        event, self.root = next(self.context)

        
    def __iter__(self):
        return(self)

    
    def __next__(self):
        for event, elem in self.context:
            if event == "end" and elem.tag == "BioSample":
                bios = BioSample()
                bios['is_reference']=None
                for k,v in elem.items():
                    bios[k] = v
                bios['id_recs'] = []
                bios['ids'] = []
                bios['sra_sample']=None
                bios['dbgap']=None
                bios['gsm']=None
                for id in elem.iterfind('.//Id'):
                    idrec = {'db': id.get('db'), 'label': id.get('db_label'), 'id':id.text}
                    bios['ids'].append(idrec['id'])
                    bios['id_recs'].append(idrec)
                    # add xref fields for SRA, dbGaP, and GEO
                    
                    if(id.get('db')=='SRA'):
                        bios['sra_sample']=id.text
                    if(id.get('db')=='dbGaP'):
                        bios['dbgap']=id.text
                    if(id.get('db')=='GEO'):
                        bios['gsm']=id.text
                bios['title'] = elem.findtext('.//Description/Title')
                bios['description'] = elem.findtext('.//Description/Comment/Paragraph')
                organism = elem.find('.//Organism')
                bios['taxonomy_name'] = organism.get('taxonomy_name')
                bios['taxon_id'] = int(organism.get('taxonomy_id'))
                bios['attribute_recs'] = []
                bios['attributes'] = []
                for attribute in elem.findall('.//Attribute'):
                    attrec = attribute.attrib
                    attrec['value'] = attribute.text
                    bios['attribute_recs'].append(attrec)
                    try:
                        bios['attributes'].append(attrec['harmonized_name'])
                    except:
                        bios['attributes'].append(attrec['attribute_name'])
                        
                bios['model'] = elem.findtext('.//Model')
                #print(json.dumps(bios))
                #res = es.index(index="bioes", doc_type='biosample', id=bios['id'], body=bios)
                elem.clear()
                return bios



import omicidx.geo.parser as op
import click
import json

@click.group()
def cli():
    pass


@cli.command()
@click.option('--gse',
              help='A single GSE accession')
def gse_to_json(gse):
    print(op.gse_to_json(gse))

@cli.command()
@click.option('--term',
              help='something like "2019/01/28:2019/02/28[PDAT]"')
@click.option("--outfile", type=click.File('w'),
              help="output file name, default is stdout")
def bulk_gse_to_json(term, outfile = '/dev/stdout'):
    op.bulk_gse_to_json(term, outfile)


@cli.command()
@click.option('--term',
              help='something like "2019/01/28:2019/02/28[PDAT]"')
@click.option("--outfile",
              help="output file name, default is stdout")
def gse_accessions(term, outfile = '/dev/stdout'):
    op.gse_accessions(term, outfile)
    
if __name__ == '__main__':
    cli()



"""Simple utility functions for EBI ENA"""
import urllib.request
import xml.etree.ElementTree as etree
import csv
import io
import omicidx.sra_parsers

class EBIEna(object):
    def __init__(self):
        pass

    def get_xml_for_accession(self, accession):
        """Get an XML record for a single accession as an ElementTree Node from EBI

        Parameters
        ----------
        accession: str
            Any SRA/ENA accession

        Returns
        -------
        an etree ElementNode parsed from the retrieved XML."""
        url = "https://www.ebi.ac.uk/ena/data/view/{accession}&display=xml&download=xml&filename={accession}.xml"
        url = url.format(accession = accession)
        with urllib.request.urlopen(url) as response:
            return(etree.parse(response).getroot()[0])

    def get_file_report_for_accession(self, accession):
        """Get a file report from EBI

        Parameters
        ----------
        accession: str
            Any SRA/ENA accession

        Returns
        -------
        a list of rows representing available files for the accession
        """
        url = "https://www.ebi.ac.uk/ena/data/warehouse/filereport?accession={accession}&result=read_run"
        url = url.format(accession = accession)
        print(url)
        response = urllib.request.urlopen(url)
        reader = csv.DictReader(io.StringIO(response.read().decode()),delimiter="\t")
        return(list(reader))

    def get_parsed_xml_for_accession(self, accession):
        xml = self.get_xml_for_accession(accession)
        xmlfunc = getattr(omicidx.sra_parsers,'{}_xml_iter_parser'.format(xml.tag.lower()))
        return(xmlfunc(xml))
        
        
    



import gzip
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from typing import Tuple


def open_file(fname: str, mode: str = 'rt'):
    """open a file, including dealing with gzipped files
 
    Parameters
    ----------
    fname : str
        a filename. If ending in .gz, will use gzip.open(). 
        Otherwise, a regular open() will be used.

    mode : str
        The file opening mode. 
    """
    if(fname.endswith('.gz')):
        f = gzip.open(fname, mode)
    else:
        f = open(fname, mode)
    return(f)


def requests_retry_session(
    retries: int = 3,
    backoff_factor: float = 0.3,
    status_forcelist: Tuple[int]=(500, 502, 504),
    session: requests.Session = None,
) -> requests.Session:
    """Get a requests session that handles retries

    Parameters
    ==========
    retries: int
        Number of retries allowed
    backoff_factor: float
        Implements geometric backoff 
    status_forcelist: Tuple[int]
        Status codes to retry for
    session: requests.Session 
        Supply a session that will be modified, or 
        a new session will be created.
        
    Returns
    =======
    A new requests.Session with retries enforced.
    ```
t0 = time.time()
try:
    response = requests_retry_session().get(
        'http://httpbin.org/delay/10',
        timeout=5
    )
except Exception as x:
    print('It failed :(', x.__class__.__name__)
else:
    print('It eventually worked', response.status_code)
finally:
    t1 = time.time()
    print('Took', t1 - t0, 'seconds')
    ```
    """

    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omicidx',
 'omicidx.db',
 'omicidx.geo',
 'omicidx.schema',
 'omicidx.scripts',
 'omicidx.sra']

package_data = \
{'': ['*']}

install_requires = \
['Click',
 'biopython==1.75',
 'boto3>=1.9,<2.0',
 'pydantic',
 'requests>=2.22,<3.0',
 'sd_cloud_utils',
 'sqlalchemy>=1.3,<2.0',
 'ujson>=1.35,<2.0']

entry_points = \
{'console_scripts': ['omicidx_tool = omicidx.scripts.cli:cli']}

setup_kwargs = {
    'name': 'omicidx',
    'version': '0.3.16',
    'description': 'The OmicIDX project collects, reprocesses, and then republishes metadata from multiple public genomics repositories. Included are the NCBI SRA, Biosample, and GEO databases. Publication is via the cloud data warehouse platform Bigquery, a set of performant search and retrieval APIs, and a set of json-format files for easy incorporation into other projects.',
    'long_description': '#\n\n# New process\n\n\n## Steps\n\n- Download xml\n- Create basic json\n- Upload json to s3\n- munge basic json to parquet\n- munge parquet to \n    - experiment joined\n\t- sample joined\n\t- run joined\n\t- study with aggregates\n\t- Include aggs in spark jobs:\n\t\t- number of samples, experiments, runs\n\t\t- sample, experiment, and run accessions (as array)\n- Save munged spark data (json, parquet)\n- Create elasticsearch index mappings\n- Drop existing elasticsearch mappings\n- Load elasticsearch index mappings\n\n\n## lambda\n\nzip lambdas.zip lambda_handlers.py sra_parsers.py\n\n\naws lambda create-function --function-name sra_to_json --zip-file fileb://lambdas.zip --handler lambda_handlers.lambda_return_full_experiment_json --runtime python3.6 --role arn:aws:iam::377200973048:role/lambda_s3_exec_role\n\n\n# Invoke\n\naws lambda invoke --function-name sra_to_json --log-type Tail --payload \'{"accession":"SRX000273"}\' /tmp/abc.txt\n\n# Concurrency\n\n1000 total, reserve for certain functions to limit, etc.\n\naws lambda put-function-concurrency --function-name sra_to_json --reserved-concurrent-executions 20\n\n# timeout and memory\n\naws lambda update-function-configuration --function-name sra_to_json --timeout 15\n\n\n# logging\n\nhttps://github.com/jorgebastida/awslogs\n\nawslogs get /aws/lambda/sra_to_json ALL --watch\n\n\n## dynamodb\n\naws dynamodb scan --table-name sra_experiment --select "COUNT"\n\n# GEO\n\n```\npython -m omicidx.geometa --gse=GSE10\n```\n\nWill print json, one "line" per entity to stdout.\n\n',
    'author': 'Sean Davis',
    'author_email': 'seandavi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/omicidx/omicidx-parsers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
